`timescale 10ns / 10ps

`define VSYNCSTART			0
`define VSYNCEND				2
`define HSYNCSTART			0
`define HSYNCEND				96
`define HTOTAL					800
`define HBACKPORCH			64
`define VTOTAL					525
`define HACTSTART				144
`define HACTEND				784
`define VACTSTART				35
`define VACTEND				515
`define VBACKPORCH			48
`define VACTPAPER				467

module aZXSpectrum
(
	input 				key1,							// Debug key
	input 				key2,							// Debug key
	input           	clk,           			// system clock 50Mhz on board
	input           	rst_n,         			// reset, low active
	output reg[3:0] 	led,           			// LED, use for control the LED signal on board
	output  				vga_hsync,					// Out h sync
	output  				vga_vsync,    				// Out v sync
	output reg[4:0]	out_R,						// Output RED
	output reg[5:0]	out_G,						// Output GREEN
	output reg[4:0]	out_B							// Output BLUE
);

reg		  outVideoPCLK;							// VGA pclk counter
reg		  [9:0]outPixCntr;						// Pixels in line counter
reg		  [9:0]outLineCntr;						// Lines in frame counter
reg		  outVsync;									// Output vsync
reg 		  outHsync;									// Output hsync
reg [23:0] pixeladdr;								// Pixel address before transferring into ZX format
reg [13:0] aVRAM_addr_a;							// Video memory address - for video adapter
reg [13:0] aVRAM_addr_b;							// Video memory address - for Z80
reg [7:0]  aVRAM_data_a;							// Video data IN - for video adapter, not used
reg [7:0]  aVRAM_data_b;							// Video data IN - for Z80
reg [6:0]  ColorLineAddr;
reg [255:0]  ColorDataMemCpy;
reg [7:0] CurColorAttribute;
reg [2:0]  TempColor;
reg 		  ColorSync;
wire [7:0] aVRAM_q_a;								// Video data OUT - for video adapter		
wire [7:0] aVRAM_q_b;								// Video data OUT - for Z80
reg aVRAM_wren_a;										// Video data, Write enable - for video adapter
reg aVRAM_wren_b;										// Video data, Write enable - for Z80

assign vga_hsync = outHsync;						// Assign output hsync pin to signal name
assign vga_vsync = outVsync;        			// Assign output vsync pin to signal name

 
always@(posedge clk or negedge rst_n)			// Main process for video
begin
	if (rst_n == 1'b0)								// If reset is active
	begin
		outVideoPCLK <= 0; 							// Set PCLK out to 0
		led[3:0]	<= 0;									// Set led to 0	
		outPixCntr <= 0;								// Set Pixels in line counter to 0
		outVsync <= 0;									// Set output vsync to 0
		outHsync <= 0;									// Set output hsync to 0
		outLineCntr <= 0;								// Set lines in frame counter to 0
		out_B <= 0;										// Set B default to 0
		out_G <= 0;										// Set G default to 0
		out_R <= 0;										// Set R default to 0
		aVRAM_data_a <= 0;							// Set video data in from videoadater to 0
		aVRAM_wren_a <= 0;							// Set write enable from videoadater to 0
		aVRAM_addr_a <= 0;							// Set video adatper address to 0	
		pixeladdr <= 0;								// Set plain pixel address (not in ZX way) to 0
		ColorLineAddr <= 0;
		ColorSync = 0;
		TempColor = 0;
		ColorDataMemCpy = 0;
		CurColorAttribute = 0;
	end else												// if NOT reset
	begin												
		if (outVideoPCLK == 1)						// We need to devide 50M osc by 2 to get to 25M and assign active edge, so it will be H->L
		begin
			outVideoPCLK <= 0;						// Clear output PCLK
			led[3:0] <= 0;								// Set LED to 0
			
			case (outLineCntr)						// Check if it is time for Vsync
				`VSYNCSTART: begin					// If we are at the beggining of the frame then set Vsync to 0 as polarity is NEG
					outVsync <= 0;
				end
				`VSYNCEND: begin						// If we are at the end of the frame sync pulse then set Vsync to 1
					outVsync <= 1;
				end
			endcase
															// If we are in the active frame
			if (outPixCntr >= `HACTSTART && outPixCntr < `HACTEND && outLineCntr >= `VACTSTART && outLineCntr < `VACTEND)
			begin
															// ZX Spectrum screen is 256x192px, double that and it gives us 512x324 px
															// Hbackporch = (640-512)/2 = 64; Vbackporch = (480-324)/2 = 78
															// If we are at the edge the draw while rectangle  35+78  till 515 - 78  ie 384
				if (outPixCntr >= `HACTSTART + `HBACKPORCH && outPixCntr < `HACTEND - `HBACKPORCH && outLineCntr >= `VACTSTART + `VBACKPORCH && outLineCntr < `VACTPAPER)
				begin
					if (aVRAM_q_a[7 - outPixCntr[3:1]] == 1)
					//if (1 == 1)
					begin
						if (outPixCntr[3:0] == 0)
						begin
							CurColorAttribute = ColorDataMemCpy[255:248];
							ColorDataMemCpy = ColorDataMemCpy << 8;
						end
						
						if ((CurColorAttribute & 1) > 0)
						begin
							out_B <= 5'b11111;
						end else
						begin
							out_B <= 0;			
						end
						
						if ((CurColorAttribute & 2) > 0)
						begin
							out_G <= 6'b111111;			
						end else
						begin
							out_G <= 0;			
						end
						
						if ((CurColorAttribute & 4) > 0)
						begin
							out_R <= 5'b11111;			// R = max red
						end else
						begin
							out_R <= 0;
						end
					end else
					begin
						if (outPixCntr[3:0] == 0)
						begin
							CurColorAttribute = ColorDataMemCpy[255:248];
							ColorDataMemCpy = ColorDataMemCpy << 8;
						end

						out_B <= 5'b00000;			// B = max blue
						out_G <= 6'b000000;			// G = max green
						out_R <= 5'b00000;			// R = max red
					end
				end else
				begin
															//Border color
					out_B <= 5'b11111;						
					out_G <= 0;
					out_R <= 0;
				end
			end else
			begin
					out_B <= 0;							//If we are in the balnking area then assign B = 0
					out_G <= 0;							//If we are in the balnking area then assign G = 0
					out_R <= 0;							//If we are in the balnking area then assign R = 0
					
					if (outLineCntr >= `VACTSTART + `VBACKPORCH && outLineCntr < `VACTPAPER && outPixCntr < `HACTSTART + `HBACKPORCH)
					begin
						if (ColorSync == 1)
						begin
							ColorDataMemCpy[7:0] = 2;//aVRAM_q_a; //ColorLineAddr;
							if (ColorLineAddr < 31)
							begin
								ColorDataMemCpy = ColorDataMemCpy << 8;
							end
						end 
					end else
					begin
						ColorDataMemCpy = 0;
					end
			end											
			
			if (outLineCntr >= `VACTSTART + `VBACKPORCH && outLineCntr < `VACTPAPER && outPixCntr < `HACTSTART + `HBACKPORCH - 1)
			begin
				if (ColorSync == 0)
				begin
					TempColor = TempColor + 1;
					aVRAM_addr_a = 6144 + ColorLineAddr + (((outLineCntr - `VACTSTART - `VBACKPORCH) >> 4)*32);
					if (ColorLineAddr < 31)
					begin
						ColorLineAddr = ColorLineAddr + 1;
					end
				end
			end else
			begin
				ColorLineAddr = 0;
				TempColor = 0;
			end
															//If NEXT pixel is in active area then setup address to read it
			if (outPixCntr >= `HACTSTART + `HBACKPORCH - 1 && outPixCntr < `HACTEND - `HBACKPORCH - 1 && outLineCntr >= `VACTSTART + `VBACKPORCH && outLineCntr < `VACTPAPER)
			begin
				pixeladdr = (((outPixCntr + 1 - `HACTSTART - `HBACKPORCH)>>1) >> 3) + (((outLineCntr - `VACTSTART - `VBACKPORCH) >> 1) * 32); 
															// Blocking assignments literally block the execution of the next
															// statement until the current statement is executed. (blocking '=', non blocking '<=')
															// – Consequently, blocking assignments result in ordered
															// statement execution.
															// For example:
															// assume a = b = 0 initially;
															// a = 1; //executed first
															// b = a; //executed second
															// then a = 1, b = 1 after ordered execution. So:
															// ZX video memory is organized as following:
															// Y7  Y6  Y2  Y1  Y0 Y5  Y4 Y3 X4 X3 X2 X1 X0 ..or..
															// Y12 Y11 Y7  Y6  Y5 Y10 Y9 Y8 X4 X3 X2 X1 X0
				aVRAM_addr_a = pixeladdr[4:0] | pixeladdr[10:8] << 5 | pixeladdr[7:5] << 8 | pixeladdr [12:11] << 11;
			end

			case (outPixCntr)							// Check if it is time for Hsync
				`HSYNCSTART: begin					// If we are at the beginning of the line then set Hsync to 0 as polarity is NEG
					outHsync <= 0;
				end
				`HSYNCEND: begin						// If we are at the end of horizontal sync pulse then set Hsync to 1
					outHsync <= 1;
				end
			endcase
			
		end else											// If we are at the L->H PCLK satate
		begin
			ColorSync = ~ColorSync;
			outVideoPCLK = 1;							// Assign 1 to PCLK
			led[3:0] <= 1;								// Set LED to 1
			if (outPixCntr >= `HTOTAL - 1)		// If we are done with the line	
			begin
				outPixCntr <= 0;						// Set H pixel count to 0
				if (outLineCntr >= `VTOTAL - 1)	// If we are done with the frame
				begin
					outLineCntr <= 0;					// Set V counter to 0
				end else
				begin										// If not done then increment it
					outLineCntr <= outLineCntr + 1;
				end
			end else
			begin
				outPixCntr <= outPixCntr + 1;		// If we are not done with the line then increment it
			end
		end
	end
end

endmodule

